"""Business logic default values"""

# Minimum spanning reads pair per bilion required for non a non-IGH
# DUX4-rearrangement to be called
srpb_threshold = 20.0
# Minimum mapping quality required outside of DUX4 and IGH regions to be
# counted as spanning read
minimum_mapq = 10
